"""
Dasladen ETL
A Simple, Tiny and Ridiculus ETL Engine made with Python
"""

import os
import sys

from argparse import ArgumentParser
from schedule import run_pending, every
from time import sleep
from shutil import copy
from processor import Watcher
from log import add_log_handler, ConsoleHandler


def main():
    parser = ArgumentParser(description="DasLaden ETL")
    parser.add_argument("-capture", default="capture", help="Capture folder")
    parser.add_argument("--verbose", nargs="?", default=False, const=True, help="Output logs to console")
    args = parser.parse_args()
    v = vars(args)

    if v["verbose"]:
        add_log_handler(ConsoleHandler())

    print ("DasLaden ETL started. (Press CTRL+C to stop)")
    watch = Watcher(v["capture"])

    if os.path.isfile("./start.zip"):
        copy('./start.zip', v["capture"])

    # make path for dynamic module import
    sys.path.append('{}/module'.format(os.getcwd()))

    # TODO: receive a parameter for watch time
    every(10).seconds.do(watch.check)

    while True:
        try:
            run_pending()
            sleep(1)
        except KeyboardInterrupt:
            print ("\nDasLaden ETL finished.")
            break


if __name__ == "__main__":
    main()





"""
Task Module

External Dependencies:
- petl : (c) 2012 Alistair Miles - MIT License (https://pypi.python.org/pypi/petl/1.1.0)

Features:
- CSV -> DB task
- DB -> CSV task
- DB -> DB task
- CSV -> CSV task
- XLS -> CSV task
- XML -> DB task
- XML -> CSV task
- FTP Upload task
- ZIP task
- Python Module task
- SQL task

"""

import os
import sys
import petl as etl
import ftputil
import zipfile
import importlib

from string import maketrans
from log import get_time_filename
from taskdriver import *


class Connection(object):
    def __init__(self, config):
        self._connections = config.get("connections", None)
        
    def get_connection(self, name):
        if self._connections is not None:
            for conn in self._connections:
                if conn["name"] == name:
                    return conn
        raise KeyError()


class DriverFactory(object):
    def __init__(self, config):
        self._connections = Connection(config)

    def get_connection(self, name):
        return self._connections.get_connection(name)

    def get_driver(self, name):
        item = self.get_connection(name)
        # setup environment configuration before open connection
        if "environment" in item:
            for env in item["environment"]:
                key = env["key"]
                key = key.encode('utf-8') if isinstance(key, unicode) else key
                value = env["value"]
                value = value.encode('utf-8') if isinstance(value, unicode) else value
                os.environ[key] = value
        # select driver 
        if item["driver"] == "MySQL":
            return MySQLDriver(item)
        elif item["driver"] == "Oracle":
            return OracleDriver(item)
        elif item["driver"] == "MSSQL":
            return MSSQLDriver(item)
        
        raise NotImplementedError    


class BaseTask(object):
    """Base class for tasks"""

    def run(self, driver, task, log):
        """Run Forrest, run
        :param driver: driver factory
        :param task: task configuration
        :param log: logger object
        """
        raise NotImplementedError("implement it")

    def _parse_sql(self, task_node):
        sql = u''
        if isinstance(task_node, dict):
            if "command" in task_node:
                sql = task_node["command"]
            elif "query" in task_node:
                path = task_node.get("path", "input")
                input_query = u"{}/{}".format(path, task_node["query"])
                with open(input_query, "r") as q:
                    sql = reduce(lambda x, y: u"{} {}".format(x, y), q.readlines())
            sql = sql.format(**task_node["params"]) if "params" in task_node else sql

        sql = sql[:-1] if sql.endswith(";") else sql
        return sql


class TransformSubTask(object):
    def __init__(self, task, log):
        self.task = task
        self.log = log

    def _modules_transform(self, record_set):
        if "transforms" in self.task:
            for transform in self.task["transforms"]:
                record_set = self._module_transform(record_set, transform)
        elif "transform" in self.task:
            record_set = self._module_transform(record_set, self.task["transform"])

        return record_set

    # noinspection PyMethodMayBeStatic
    def _module_transform(self, record_set, transform=None):
        if transform is not None:
            module_name = transform["module"]
            package = transform.get("package", None)
            module_obj = importlib.import_module(module_name, package)
            reload(module_obj)
            fields = transform.get("fields", [])
            args = transform.get("args", {})
            if "class" in transform:
                self.log.write(u"Transform data with {}".format(transform["class"]))
                instance = getattr(module_obj, transform["class"])(**args)
                record_set = instance.transform(record_set, *fields)
            else:
                record_set = module_obj.transform(record_set, *fields, **args)

        return record_set

    def _petl_transform(self, record_set):
        if "transform" in self.task:
            transform = self.task["transform"]
            if "convert" in transform:
                conversions = {}
                for field, func in transform["convert"]:
                    conversions[field] = func
                record_set = etl.convert(record_set, conversions)

            if "filter" in transform:
                record_set = etl.select(record_set, transform["filter"])

            if "remove" in transform:
                cuts = []
                for field in transform["remove"]:
                    cuts.append(field)
                record_set = etl.cutout(record_set, cuts)

            if "rename" in transform:
                names = {}
                for old, new_one in transform["rename"]:
                    names[old] = new_one
                record_set = etl.rename(record_set, names)

        return record_set

    def get_result(self, record_set):
        record_set = self._modules_transform(record_set)
        record_set = self._petl_transform(record_set)
        return record_set


class DbCsvTask(BaseTask):

    def run(self, driver, task, log):
        input_driver = driver.get_driver(task["source"]["connection"])
        sql = self._parse_sql(task["source"])
        db = input_driver.get_db()
        record_set = etl.fromdb(db, sql)
        if not etl.data(record_set).any():
            log.write("Task skipped. No rows on source")
        else:
            transform = TransformSubTask(task, log)
            record_set = transform.get_result(record_set)

            fld = task["target"].get("folder", "output")
            fld = fld.encode('latin1') if isinstance(fld, unicode) else fld
            target = task["target"]["file"]
            target = target.encode('latin1') if isinstance(target, unicode) else target
            out = "{}/{}".format(fld, target)

            separator = task["target"].get("delimiter", ";")
            separator = separator.encode('utf-8') if isinstance(separator, unicode) else separator
            enc = task["target"].get("encoding", "utf-8")

            task_log = "log/db-csv_{}_{}.log".format(task["name"], get_time_filename())
            with open(task_log, "w") as lg:
                if "truncate" in task["target"] and task["target"]["truncate"]:
                    record_set.progress(10000, out=lg).tocsv(out, encoding=enc, delimiter=separator)
                else:
                    record_set.progress(10000, out=lg).appendcsv(out, encoding=enc, delimiter=separator)
        db.close()


class CsvDbTask(BaseTask):

    def run(self, driver, task, log):
        source_folder = task["source"].get("folder", "input")
        source_folder = source_folder.encode('latin1') if isinstance(source_folder, unicode) else source_folder
        source = task["source"]["file"]
        source = source.encode('latin1') if isinstance(source, unicode) else source
        inp = "{}/{}".format(source_folder, source)

        separator = task["source"].get("delimiter", ";")
        separator = separator.encode('utf-8') if isinstance(separator, unicode) else separator

        enc = task["source"].get("encoding", "utf-8")
        enc = enc.encode('utf-8') if isinstance(enc, unicode) else enc

        record_set = etl.fromcsv(inp, encoding=enc, delimiter=separator)

        if not etl.data(record_set).any():
            log.write("Task skipped. No rows on source")
        else:
            transform = TransformSubTask(task, log)
            record_set = transform.get_result(record_set)

            output_driver = driver.get_driver(task["target"]["connection"])
            db = output_driver.get_db()

            table = task["target"]["table"]
            table = table.encode('utf-8') if isinstance(table, unicode) else table
            if "schema" in task["target"]:
                schema_name = task["target"]["schema"]
                schema_name = schema_name.encode('utf-8') if isinstance(schema_name, unicode) else schema_name
            else:
                schema_name = None

            task_log = "log/csv-db_{}_{}.log".format(task["name"], get_time_filename())
            with open(task_log, "w") as lg:
                if "truncate" in task["target"] and task["target"]["truncate"]:
                    record_set.progress(10000, out=lg).todb(output_driver.cursor(db), tablename=table, schema=schema_name)
                else:
                    record_set.progress(10000, out=lg).appenddb(output_driver.cursor(db), tablename=table, schema=schema_name)

            db.close()


class DbDbTask(BaseTask):

    def run(self, driver, task, log):
        input_driver = driver.get_driver(task["source"]["connection"])
        sql = self._parse_sql(task["source"])
        db = input_driver.get_db()
        record_set = etl.fromdb(db, sql)
        if not etl.data(record_set).any():
            log.write("Task skipped. No rows on source")
        else:
            transform = TransformSubTask(task, log)
            record_set = transform.get_result(record_set)

            output_driver = driver.get_driver(task["target"]["connection"])
            out_db = output_driver.get_db()

            table = task["target"]["table"]
            table = table.encode('utf-8') if isinstance(table, unicode) else table
            if "schema" in task["target"]:
                schema_name = task["target"]["schema"]
                schema_name = schema_name.encode('utf-8') if isinstance(schema_name, unicode) else schema_name
            else:
                schema_name = None

            task_log = "log/db-db_{}_{}.log".format(task["name"], get_time_filename())
            with open(task_log, "w") as lg:
                if "truncate" in task["target"] and task["target"]["truncate"]:
                    record_set.progress(10000, out=lg).todb(output_driver.cursor(out_db), tablename=table, schema=schema_name)
                else:
                    record_set.progress(10000, out=lg).appenddb(output_driver.cursor(out_db), tablename=table, schema=schema_name)

            out_db.close()
        db.close()


class CsvCsvTask(BaseTask):

    def run(self, driver, task, log):
        inp = task["source"]["file"]
        inp = inp.encode('latin1') if isinstance(inp, unicode) else inp
        inp = "input/{}".format(inp)
        separator = task["source"].get("delimiter", ";")
        separator = separator.encode('utf-8') if isinstance(separator, unicode) else separator

        enc = task["source"].get("encoding", "utf-8")
        enc = enc.encode('utf-8') if isinstance(enc, unicode) else enc
        
        record_set = etl.fromcsv(inp, encoding=enc, delimiter=separator)
        if not etl.data(record_set).any():
            log.write("Task skipped. No rows on source")
        else:
            transform = TransformSubTask(task, log)
            record_set = transform.get_result(record_set)

            out = task["target"]["file"]
            out = out.encode('latin1') if isinstance(out, unicode) else out
            out = "output/{}".format(out)
            separator = task["target"].get("delimiter", ";")
            separator = separator.encode('utf-8') if isinstance(separator, unicode) else separator
            enc = task["target"].get("encoding", "utf-8")

            task_log = "log/csv-csv_{}_{}.log".format(task["name"], get_time_filename())
            with open(task_log, "w") as lg:
                if "truncate" in task["target"] and task["target"]["truncate"]:
                    record_set.progress(10000, out=lg).tocsv(out, encoding=enc, delimiter=separator)
                else:
                    record_set.progress(10000, out=lg).appendcsv(out, encoding=enc, delimiter=separator)


class XlsCsvTask(BaseTask):

    def run(self, driver, task, log):
        inp = task["source"]["file"]
        inp = inp.encode('latin1') if isinstance(inp, unicode) else inp
        inp = "input/{}".format(inp)
        sheet = task["source"].get("sheet", None)
        use_view = task["source"].get("use_view", True)
       
        record_set = etl.fromxls(inp, sheet, use_view=use_view)
        if not etl.data(record_set).any():
            log.write("Task skipped. No rows on source")
        else:
            transform = TransformSubTask(task, log)
            record_set = transform.get_result(record_set)

            out = task["target"]["file"]
            out = out.encode('latin1') if isinstance(out, unicode) else out
            out = "output/{}".format(out)
            separator = task["target"].get("delimiter", ";")
            separator = separator.encode('utf-8') if isinstance(separator, unicode) else separator
            enc = task["target"].get("encoding", "utf-8")

            task_log = "log/xls-csv_{}_{}.log".format(task["name"], get_time_filename())
            with open(task_log, "w") as lg:
                if "truncate" in task["target"] and task["target"]["truncate"]:
                    record_set.progress(10000, out=lg).tocsv(out, encoding=enc, delimiter=separator)
                else:
                    record_set.progress(10000, out=lg).appendcsv(out, encoding=enc, delimiter=separator)


class XmlCsvTask(BaseTask):

    def run(self, driver, task, log):
        inp = task["source"]["file"]
        inp = inp.encode('latin1') if isinstance(inp, unicode) else inp
        inp = "input/{}".format(inp)
        row_match = task["source"].get("row", None)
        value_match = task["source"].get("value", None)
        attr = task["source"].get("attr", None)
        mapping = task["source"].get("mapping", None)

        if row_match and value_match:
            if attr:
                record_set = etl.fromxml(inp, row_match, value_match, attr)
            else:
                record_set = etl.fromxml(inp, row_match, value_match)
        elif row_match and mapping:
            record_set = etl.fromxml(inp, row_match, mapping)
        else:
            raise ValueError('Incorrect parameter for source')

        if not etl.data(record_set).any():
            log.write("Task skipped. No rows on source")
        else:
            transform = TransformSubTask(task, log)
            record_set = transform.get_result(record_set)

            out = task["target"]["file"]
            out = out.encode('latin1') if isinstance(out, unicode) else out
            out = "output/{}".format(out)
            separator = task["target"].get("delimiter", ";")
            separator = separator.encode('utf-8') if isinstance(separator, unicode) else separator
            enc = task["target"].get("encoding", "utf-8")

            task_log = "log/xml-csv_{}_{}.log".format(task["name"], get_time_filename())
            with open(task_log, "w") as lg:
                if "truncate" in task["target"] and task["target"]["truncate"]:
                    record_set.progress(10000, out=lg).tocsv(out, encoding=enc, delimiter=separator)
                else:
                    record_set.progress(10000, out=lg).appendcsv(out, encoding=enc, delimiter=separator)


class XmlDbTask(BaseTask):

    def run(self, driver, task, log):
        inp = task["source"]["file"]
        inp = inp.encode('latin1') if isinstance(inp, unicode) else inp
        inp = "input/{}".format(inp)
        row_match = task["source"].get("row", None)
        value_match = task["source"].get("value", None)
        attr = task["source"].get("attr", None)
        mapping = task["source"].get("mapping", None)

        if row_match and value_match:
            if attr:
                record_set = etl.fromxml(inp, row_match, value_match, attr)
            else:
                record_set = etl.fromxml(inp, row_match, value_match)
        elif row_match and mapping:
            record_set = etl.fromxml(inp, row_match, mapping)
        else:
            raise ValueError('Incorrect parameter for source')

        if not etl.data(record_set).any():
            log.write("Task skipped. No rows on source")
        else:
            transform = TransformSubTask(task, log)
            record_set = transform.get_result(record_set)

            output_driver = driver.get_driver(task["target"]["connection"])
            db = output_driver.get_db()

            table = task["target"]["table"]
            table = table.encode('latin1') if isinstance(table, unicode) else table
            if "schema" in task["target"]:
                schema_name = task["target"]["schema"]
                schema_name = schema_name.encode('latin1') if isinstance(schema_name, unicode) else schema_name
            else:
                schema_name = None

            task_log = "log/xml-db_{}_{}.log".format(task["name"], get_time_filename())
            with open(task_log, "w") as lg:
                if "truncate" in task["target"] and task["target"]["truncate"]:
                    record_set.progress(10000, out=lg).todb(output_driver.cursor(db),
                                                            tablename=table, schema=schema_name)
                else:
                    record_set.progress(10000, out=lg).appenddb(output_driver.cursor(db),
                                                                tablename=table, schema=schema_name)
            db.close()


class FtpUploadTask(BaseTask):

    def run(self, driver, task, log):
        source = task["source"]["file"]
        source = source.encode('latin1') if isinstance(source, unicode) else source
        source_path = task["source"].get("path", "output")
        source_path = source_path.encode('latin1') if isinstance(source_path, unicode) else source_path
        source_path = "{}/{}".format(source_path, source)
        target = task["target"].get("file", source)
        target_path = "{}/{}".format(task["target"]["path"], target)
        
        item = driver.get_connection(task["target"]["connection"])
        with ftputil.FTPHost(item["host"],
                             item["user"],
                             item["pass"]) as host:
            host.upload_if_newer(source_path, target_path)
            

class ZipTask(BaseTask):

    @staticmethod
    def _encode_cp437(s):
        return s.encode('cp437', errors='replace').translate(maketrans('?', '_'))

    def run(self, driver, task, log):
        source = task["source"]["files"]
        source_path = task["source"].get("path", "output")
        source_path = source_path.encode('latin1') if isinstance(source_path, unicode) else source_path
        remove_after = task["source"].get("remove_after", [])
        if "target" in task:
            target = task["target"]["file"] if "file" in task["target"] else "{}.zip".format(source[0])
            target_path = task["target"].get("path", source_path)
        else:
            target = "{}.zip".format(source[0])
            target_path = source_path
        target = target.encode('latin1') if isinstance(target, unicode) else target
        target = "{}.zip".format(target) if not target.endswith(".zip") else target
        target_path = target_path.encode('latin1') if isinstance(target_path, unicode) else target_path

        target_file = "{}/{}".format(target_path, target)
        with zipfile.ZipFile(target_file, 'w', zipfile.ZIP_DEFLATED) as z:
            for file_name in source:
                file_name = file_name.encode('latin1') if isinstance(file_name, unicode) else file_name
                z.write("{}/{}".format(target_path, file_name), ZipTask._encode_cp437(file_name))

        for file_name in remove_after:
            os.remove("{}/{}".format(target_path, file_name))


class PyExecTask(BaseTask):

    def run(self, driver, task, log):
        module_name = task["source"]["module"]
        package = task["source"].get("package", None)
        if "args" in task["source"]:
            args = task["source"]["args"]
            sys.argv[1:] = args
        else:
            sys.argv[1:] = []

        module_obj = importlib.import_module(module_name, package)
        reload(module_obj)
        # TODO: better no call main
        module_obj.main()


class SqlExecTask(BaseTask):

    def run(self, driver, task, log):
        output_driver = driver.get_driver(task["target"]["connection"])
        db = output_driver.get_db()
        sql = self._parse_sql(task["source"])
        cur = output_driver.cursor(db)
        cur.execute(sql)
        db.commit()
        db.close()


class NopTask(BaseTask):
    """Nop task to user programming json and disable a task without remove from task file"""
    def run(self, driver, task, log):
        log.write(u'Nothing to do. Disabled task? Check it John Snow!')


class CustomTask(BaseTask):

    def run(self, driver, task, log):
        log.write(u"Loading custom task.")
        module_name = task["module"]
        package = task.get("package", None)
        module_obj = importlib.import_module(module_name, package)
        reload(module_obj)
        task_class = getattr(module_obj, task["class"])
        task_instance = task_class()
        task_instance.run(driver, task, log)


class TaskFactory(object):
    _tasks = {
        "db-csv": DbCsvTask,
        "csv-db": CsvDbTask,
        "db-db": DbDbTask,
        "csv-csv": CsvCsvTask,
        "xls-csv": XlsCsvTask,
        "xml-csv": XmlCsvTask,
        "xml-db": XmlDbTask,
        "ftp-upload": FtpUploadTask,
        "zip": ZipTask,
        "py-exec": PyExecTask,
        "sql-exec": SqlExecTask,
        "nop": NopTask,
        "custom": CustomTask
    }

    def get_task(self, task_type):
        if task_type in self._tasks:
            return self._tasks[task_type]()
        raise NotImplementedError





"""
DasLaden ETL
A Simple, Tiny and Ridiculus ETL Engine made with Python

"""

__title__ = 'dasladen'
__author__ = 'Vagner Pagotti'
__author_email__ = 'pagotti@gmail.com'
__version__ = '0.1.1'
__license__ = 'MIT'
__copyright__ = 'Copyright (c) Vagner Pagotti'

from task import BaseTask
from processor import Watcher



"""
Processor Module
Processor classes for process files in capture folder

External Dependencies:
- schedule : (c) 2013 Daniel Bader (http://dbader.org) - MIT License (https://pypi.python.org/pypi/schedule)
- backports.tempfile: (c) Pi Delport - Python License (https://github.com/pjdelport/backports.tempfile)

Features:
- UnZip: Open zip files and process then in isolated folder
- Extract: Extract zip file tht contains task file and others files (mostly .csv) that form a task package
- Copy: Move the files to appropriated folders before execution
- Run: Open task file and checks if schedule or run the tasks
- Schedule task to later execution and recurring execution
- Capture new files on capture folder
- Process captured files

"""

import traceback
import zipfile
import schedule

from shutil import copy2
from backports import tempfile
from log import Logger, get_time_filename
from taskrun import *


class SchedulerJob(object):
    """Job information for scheduler processor"""

    def __init__(self, manager, filename, once):
        self.manager = manager
        self.once = once
        self.filename = filename

    def check(self):
        with Logger("scheduler_{}_{}".format(self.filename, get_time_filename())) as log:
            start = time.time()
            try:
                log.write("Executing Scheduled Tasks: {}".format(self.filename))
                self.manager.run(log)
            except StandardError:
                log.write("Error: {}".format(traceback.format_exc()))
            finally:
                log.write("Finished: {0}, elapsed: {1:.2f}s".format(self.filename, (time.time() - start)))
                if self.once:
                    return schedule.CancelJob


class Scheduler(object):
    """Create a schedule for late task execution based on schedule section of task file"""

    @staticmethod
    def run(job_item, job, at_time=""):
        if at_time:
            job.at(at_time).do(job_item.check)
        else:
            job.do(job_item.check)

    @staticmethod
    def enqueue(runner, filename):
        result = ""
        # get schedule info in task file
        props = runner.schedule()
        recurring = props.get("recurring", False)
        manager = TaskRunner(runner)
        job_item = SchedulerJob(manager, filename, not recurring)
        at_time = props.get("time", "")

        if recurring:
            freq = props.get("frequency", "daily")
            if freq == "daily":
                days = props.get("days", 1)
                job = schedule.every(days).day
                Scheduler.run(job_item, job, at_time)
                result = "each {} day(s)".format(days)

            elif freq == "weekly":
                weeks = props.get("weeks", 1)
                result = "each {} week(s) on {}".format(weeks, props["weekday"])

                if "monday" in props["weekday"]:
                    job = schedule.every(weeks).monday
                    Scheduler.run(job_item, job, at_time)

                if "tuesday" in props["weekday"]:
                    job = schedule.every(weeks).tuesday
                    Scheduler.run(job_item, job, at_time)

                if "wednesday" in props["weekday"]:
                    job = schedule.every(weeks).wednesday
                    Scheduler.run(job_item, job, at_time)

                if "thursday" in props["weekday"]:
                    job = schedule.every(weeks).thursday
                    Scheduler.run(job_item, job, at_time)

                if "friday" in props["weekday"]:
                    job = schedule.every(weeks).friday
                    Scheduler.run(job_item, job, at_time)

                if "saturday" in props["weekday"]:
                    job = schedule.every(weeks).saturday
                    Scheduler.run(job_item, job, at_time)

                if "sunday" in props["weekday"]:
                    job = schedule.every(weeks).sunday
                    Scheduler.run(job_item, job, at_time)

            elif freq == "minutes":
                amount = props.get("minutes", 1)
                job = schedule.every(amount).minutes
                Scheduler.run(job_item, job)
                result = "each {} minutes(s)".format(amount)

            elif freq == "hours":
                amount = props.get("hours", 1)
                job = schedule.every(amount).hours
                Scheduler.run(job_item, job)
                result = "each {} hour(s)".format(amount)

        else:
            if at_time:
                job = schedule.every().day
                Scheduler.run(job_item, job, at_time)

            else:
                return " (error) invalid schedule"  # not has a schedule

        if at_time:
            result = "{} at '{}'".format(result, at_time)
        return result


class TaskProcessor(object):
    """Main processor for task files. Take a task file and process they entries"""

    def __init__(self, files, log):
        self.log = log
        self.files = files

    def selection(self):
        return [f for f in self.files if f.endswith('.json')]

    def execute(self, path, filename):
        start = time.time()
        f = "{}/{}".format(path, filename)
        try:
            r = Runner(f)
            if r.has_schedule:
                props = r.schedule
                if "times" in props:
                    times = parse_to_int(props["times"], 0)
                    runner = TaskRunner(r)
                    current = 0
                    while current < times:
                        current += 1
                        self.log.write("Executing Tasks ({}/{}): {}".format(current, times, filename))
                        runner.run(self.log)

                elif ("infinity" in props) and props["infinity"]:
                    runner = TaskRunner(r)
                    while True:
                        self.log.write("Executing Tasks (infinity): {}".format(filename))
                        runner.run(self.log)

                else:
                    res = Scheduler.enqueue(r, filename)
                    self.log.write("Scheduling Tasks: {}, for: {}".format(filename, res))

            else:
                runner = TaskRunner(r)
                if runner:
                    self.log.write("Executing Tasks: {}".format(filename))
                    runner.run(self.log)

        except Exception:
            self.log.write("Error: {}".format(traceback.format_exc()))

        finally:
            self.log.write("Finished: {0}, elapsed: {1:.2f}s".format(filename, (time.time() - start)))
            os.remove(f)


def parse_to_int(value, fail=None):
    if isinstance(value, basestring):
        try:
            return int(value)
        except ValueError:
            return fail
    else:
        return 0


class CopyProcessor(object):
    """Copy CVS, SQL files to input folder"""

    def __init__(self, files, log):
        self.log = log
        self.files = files
        self.target_path = "input"

    def selection(self):
        return [f for f in self.files if not f.endswith('.zip')]

    def execute(self, path, filename):
        start = time.time()
        source_file = u"{}/{}".format(path, filename)
        target_path = "module" if filename.endswith('.py') else self.target_path
        target_file = u"{}/{}".format(target_path, filename)
        try:
            self.log.write("Moving File: {}".format(filename))
            if os.path.isfile(target_file):
                os.remove(target_file)
            if Runner.is_task(source_file):
                copy2(source_file, target_file)
            else:
                os.rename(source_file, target_file)
        except StandardError:
            self.log.write("Error: {}".format(traceback.format_exc()))
        finally:
            self.log.write("Finished: {0}, elapsed: {1:.2f}s".format(filename, (time.time() - start)))


class ExtractProcessor(object):
    """Extract zip files"""

    def __init__(self, files, log):
        self.log = log
        self.files = files
        self.target = ''

    def set_target(self, target):
        self.target = target

    def selection(self):
        return [f for f in self.files if f.endswith('.zip')]

    def execute(self, path, filename):
        start = time.time()
        source_file = "{}/{}".format(path, filename)
        target = self.target if self.target != '' else path
        try:
            self.log.write("Extracting: {} into '{}'".format(filename, target))
            z = zipfile.ZipFile(source_file)
            z.extractall(target)
            z.close()
            self.log.write("Removing ZIP: {}".format(filename))
            os.remove(source_file)
        except StandardError:
            self.log.write("Error: {}".format(traceback.format_exc()))
        finally:
            self.log.write("Finished: {0}, elapsed: {1:.2f}s".format(filename, (time.time() - start)))


def _process(processor, path):
    for f in processor.selection():
        processor.execute(path, f)


class ZipFilesProcessor(object):
    """Process zip files"""

    def __init__(self, files, log):
        self.log = log
        self.files = files

    def selection(self):
        return [f for f in self.files if f.endswith('.zip')]

    # noinspection PyUnusedLocal
    def execute(self, path, filename):
        extract = ExtractProcessor(self.files, self.log)
        try:
            for zip_file in extract.selection():
                with tempfile.TemporaryDirectory() as temp_folder:
                    self.log.write("Creating Temporary Dir: {}".format(temp_folder))
                    try:
                        extract.set_target(temp_folder)
                        extract.execute(path, zip_file)
                        files = [f for f in os.listdir(temp_folder)]
                        _process(CopyProcessor(files, self.log), temp_folder)
                        _process(TaskProcessor(files, self.log), temp_folder)
                    except StandardError:
                        self.log.write("Error: {}".format(traceback.format_exc()))
                    finally:
                        self.log.write("Removing Temporary Dir: {}".format(temp_folder))
        except IOError:
            self.log.write("IO Error: {}".format(traceback.format_exc()))
        except StandardError:
            self.log.write("Error: {}".format(traceback.format_exc()))


class Watcher(object):
    """Watch a folder to capture files and process task on it"""

    def __init__(self, path):
        self.path = path
        self.before = dict([(f, None) for f in os.listdir(path)])
        print(u"Watcher started on '{}'.\n".format(path))

    def _process(self, processor):
        for f in processor.selection():
            processor.execute(self.path, f)

    def process_file_list(self, file_list):
        """first unzip, then copy and finally execute tasks"""
        with Logger('watcher_{}'.format(get_time_filename())) as log:
            log.write("Starting a capture...")
            self._process(ZipFilesProcessor(file_list, log))
            self._process(CopyProcessor(file_list, log))
            self._process(TaskProcessor(file_list, log))

    def check(self):
        after = dict([(f, None) for f in os.listdir(self.path)])
        added = [f for f in after if f not in self.before]

        # on add, process files on that order
        if added:
            self.process_file_list(added)

        self.before = after





"""
Task Runner Module

External Dependencies:
- petl : (c) 2012 Alistair Miles - MIT License (https://pypi.python.org/pypi/petl/1.1.0)
- PyMySQL: (c) 2010, 2013 PyMySQL contributors - MIT License (https://pypi.python.org/pypi/PyMySQL/0.7.1)

Features:
- Wrapper to a json task file
- Facade to run the tasks

"""

import json
import os
import time

from task import TaskFactory, DriverFactory


class Runner(object):
    """Wrapper for json task file"""

    def __init__(self, task):
        if os.path.isfile(task):
            with open(task, 'r') as f:
                self._config = json.load(f)
        else:
            raise ValueError("Task file not found!")

    @staticmethod
    def is_task(filename):
        if filename.endswith(".json") and os.path.isfile(filename):
            with open(filename, 'r') as f:
                config = json.load(f)
                # TODO: maybe a property id? jsonschema?
                return "tasks" in config
        return False

    @property
    def has_schedule(self):
        if "schedule" in self._config:
            recurring = self._config["schedule"].get("recurring", False)
            return recurring
        else:
            return False

    @property
    def schedule(self):
        return self._config["schedule"]

    @property
    def config(self):
        return self._config


class TaskRunner(object):
    """Facade to run the task runner object"""

    def __init__(self, runner):
        self._config = runner.config

    def run(self, log):
        if "tasks" in self._config:
            driver = DriverFactory(self._config)
            for item in self._config["tasks"]:
                start = time.time()
                log.write(u"Executing task item: {}".format(item["name"]))
                if item.get("disabled", False):
                    task = TaskFactory().get_task("nop")
                else:
                    task = TaskFactory().get_task(item["type"])
                task.run(driver, item, log)
                log.write(u"Task item finished: {0}, time: {1:.2f}s".format(item["name"], (time.time() - start)))
            return True



"""
Driver Module

Oracle Driver External Dependencies:
- cx_oracle: (c) 2016, 2017, Oracle and/or its affiliates.
             Portions Copyright 2007-2015, Anthony Tuininga.
             Portions Copyright 2001-2007, Computronix (Canada) Ltd., Edmonton, Alberta, Canada.
             BSD License (https://github.com/oracle/python-cx_Oracle/blob/master/LICENSE.txt)

MySQL Driver External Dependencies:
- PyMySQL: (c) 2010, 2013 PyMySQL contributors - MIT License (https://pypi.python.org/pypi/PyMySQL/0.7.1)

MSSQL Driver (via ODBC) External Dependencies:
- pyodbc: (c) Michael Kleehammer - MIT License (https://github.com/mkleehammer/pyodbc)

Features:
- Connection to MS SQL
- Connection to MySQL
- Connection to Oracle

"""


class CursorProxy(object):
    """Proxy for cursor that not has support a executemany with iterators"""

    def __init__(self, cursor):
        self._cursor = cursor

    def executemany(self, statement, parameters, **kwargs):
        # convert parameters to a list
        parameters = list(parameters)
        # pass through to proxy cursor
        return self._cursor.executemany(statement, parameters, **kwargs)

    def __getattr__(self, item):
        return getattr(self._cursor, item)


try:
    import cx_Oracle as oracle
except ImportError:
    pass


class OracleDriver(object):
    """Driver for Oracle connections"""

    def __init__(self, config):
        self.config = config

    def get_db(self):
        conn = self.config
        host_address = conn.get("host", "localhost")
        port = conn.get("port", "1521")

        str_conn = "{}/{}@{}:{}/{}".format(conn["user"], conn["pass"],
                                           host_address, port, conn["service"])
        db = oracle.connect(str_conn)
        db.outputtypehandler = self.output_type_handler

        if "initializing" in conn:
            for sql in conn["initializing"]:
                db.cursor().execute(sql)

        return db

    def output_type_handler(self, cursor, name, defaultType, size, precision, scale):
        if defaultType in (STRING, FIXED_CHAR):
            return cursor.var(unicode, size, cursor.arraysize)

    def cursor(self, db):
        return CursorProxy(db.cursor())


try:
    import pyodbc as odbc
except ImportError:
    pass


class MSSQLDriver(object):
    """Driver for MS SQL connections via ODBC"""

    def __init__(self, config):
        self.config = config

    def get_db(self):
        conn = self.config
        db_charset = "CHARSET={};".format(conn["charset"]) if "charset" in conn else ""
        host_address = conn.get("host", "(local)")
        port = conn.get("port", "1433")

        if not conn["user"]:
            str_conn = ("DRIVER={{SQL Server}};SERVER={};"
                        "PORT={};DATABASE={};Trusted_Connection=yes;{}").format(
                host_address, port,
                conn["database"],
                db_charset)
        else:
            str_conn = ("DRIVER={{SQL Server}};SERVER={};"
                        "PORT={};DATABASE={};UID={};PWD={};{}").format(
                host_address, port,
                conn["database"],
                conn["user"],
                conn["pass"],
                db_charset)

        db = odbc.connect(str_conn)

        if "initializing" in conn:
            for sql in conn["initializing"]:
                db.cursor().execute(sql)
        return db

    def cursor(self, db):
        return db.cursor()


try:
    import pymysql as mysql
except ImportError:
    pass


class MySQLDriver(object):
    """Driver for MySQL connections"""

    def __init__(self, config):
        self.config = config

    def get_db(self):
        conn = self.config
        db_charset = conn.get("charset", "utf8")
        host_address = conn.get("host", "localhost")
        port = conn.get("port", 3306)

        db = mysql.connect(host=host_address,
                           port=port,
                           user=conn["user"],
                           password=conn["pass"],
                           database=conn["database"],
                           charset=db_charset,
                           local_infile=1)
        # needed for petl work correctly
        db.cursor().execute('SET SQL_MODE=ANSI_QUOTES')

        if "initializing" in conn:
            for sql in conn["initializing"]:
                db.cursor().execute(sql)
        return db

    def cursor(self, db):
        return db.cursor()




from setuptools import setup

with open("README.rst", "r") as f:
    long_description = f.read()

setup(
    name='dasladen',
    version='0.1.1',
    description='Simple, tiny and ridiculus ETL made with Python',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT License',
    packages=['dasladen'],
    author='Vagner Pagotti',
    author_email='pagotti@gmail.com',
    url='https://github.com/pagotti/dasladen',
    keywords=['etl'],
    install_requires=[
        'schedule',
        'petl',
        'backports.tempfile',
        'ftputil'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)


"""
Log Module

Features:
- Log to file
- Log to console

"""

import time


class FileHandler(object):
    def __init__(self):
        self.file = None

    def open(self, key):
        self.file = open('log/{}.log'.format(key), 'a', 0)

    def write(self, data):
        self.file.write(u"{} {}\n".format(get_time(), data))

    def close(self):
        self.file.close()


class ConsoleHandler(object):
    def open(self, key):
        pass

    # noinspection PyMethodMayBeStatic
    def write(self, data):
        print(u"{} {}".format(get_time(), data))

    def close(self):
        pass


class LogManager(object):
    def __init__(self):
        self.log_manager_instance = dict()
        self.handlers = [FileHandler()]

    def add(self, key):
        self.log_manager_instance[key] = self.handlers

    def get(self, key):
        return self.log_manager_instance[key]


_manager = LogManager()


def add_log_handler(log_handler):
    _manager.handlers.append(log_handler)


def get_time():
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())


def get_time_filename():
    return time.strftime("%Y%m%d_%H%M%S", time.localtime())


class Logger(object):
    def __init__(self, filename):
        self.key = filename
        _manager.add(filename)

    def write(self, data):
        for handler in _manager.get(self.key):
            handler.write(data)

    def __enter__(self):
        for handler in _manager.get(self.key):
            handler.open(self.key)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for handler in _manager.get(self.key):
            handler.close()


